/*

Code extending the causal state reconstruction algorithm
technique exposed in "Automatic Filters for the Detection of Coherent
Structure in Spatiotemporal Systems", by Shalizi et al.

The algorithm below can handle incremental estimation of the states as light
cones become available. Convergence is improved by running through the
state estimates in decreasing complexity. Also, states do really
handle all observations from their clustered past light cones.

The CausalStateAnalyzer class is generic and will process your custom types.
Feed it light cones incrementally as you produce them, and it will
update the causal states and return you the best state estimate so far.

User API is at the end of the class below.
See CA_Complexity.cpp for a usage example.

Nicolas Brodu, 26/09/06, file version 0.3
Released under GNU Public Licence v2 or above.

*/

#include <list>
#include <map>
// unordered_map will soon become standard C++, but it's in the Technical Report 1 namespace at the time of writing
#ifndef NO_TR1
#include <tr1/unordered_map>
#endif
#include <vector>

#include <assert.h>
#include <math.h>

template<class PastLightCone, class FutureLightCone>
struct CausalStateAnalyzer {

// IMPLEMENTATION, see user API below

    double significanceLevel;

    struct Cluster;

    /*  Holds a distribution over future cones
        It needs:
        - Sorted property
        - traversal (esp. for the match method)
        - insert/remove
        A map would provide O(n) traversal but the O is too big...
        Use a list, at the cost of O(n) search. A skip list could be worth a try.
    */
    struct FutureDistribution : public std::list< std::pair<FutureLightCone, int> > {
        int total;
        // the cluster this distribution belongs to.
        Cluster* cluster;

        typedef typename std::list< std::pair<FutureLightCone, int> >::iterator iterator;
        typedef typename std::list< std::pair<FutureLightCone, int> >::const_iterator const_iterator;

        inline FutureDistribution() : total(0), cluster(0) {}
        inline FutureDistribution(const FutureLightCone& flc) : total(0), cluster(0) {
            insert(flc);
        }

        inline int insert(const FutureLightCone& flc) {
            iterator it = this->begin();
            while (it!=this->end()) {
                if (it->first > flc) break;
                if (it->first == flc) {
                    ++it->second;
                    return ++total;
                }
                ++it;
            }
            // not found
            ((std::list< std::pair<FutureLightCone, int> >*)this)->insert(it, make_pair(flc, 1));
            return ++total;
        }

        inline int remove(const FutureLightCone& flc) {
            iterator it = this->begin();
            while (it!=this->end()) {
                if (it->first == flc) break;
                ++it;
            }
            assert(it!=this->end()); // can't remove a future that was not inserted
            if (--it->second == 0) this->erase(it); // do not keep empty entries
            return --total;
        }

        inline void add(const FutureDistribution& distribution) {
            const_iterator it2 = distribution.begin();
            const_iterator eit2 = distribution.end();
            iterator it1 = this->begin();
            iterator eit1 = this->end();
            while (it2!=eit2) {
                // catchup on distribution
                while (it1!=eit1 && it1->first < it2->first) ++it1;
                // use constant-time insert-with-hint function
                while (it2!=eit2 && (it1==eit1 || it2->first < it1->first)) {
                    ((std::list< std::pair<FutureLightCone, int> >*)this)->insert(it1, *it2++);
                }
                // now elements are equal
                while (it2!=eit2 && it1!=eit1 && it2->first == it1->first) {
                    it1->second += it2->second;
                    ++it1; ++it2;
                }
            }
            total += distribution.total;
        }

        inline void sub(const FutureDistribution& distribution) {
            const_iterator it2 = distribution.begin();
            const_iterator eit2 = distribution.end();
            iterator it1 = this->begin();
            iterator eit1 = this->end();
            while (it2!=eit2) {
                // catchup on distribution
                #ifdef NDEBUG // assume correctness
                while (it1->first < it2->first) ++it1;
                #else // assertion mode
                while (it1!=eit1 && it1->first < it2->first) ++it1;
                // user can't remove a dist that was not added: there is NO it2 not in it1, elements are equal
                assert(it1!=eit1 && it1->first == it2->first);
                #endif
                // now elements are equal
                // so, subtract, and potentially erase if count is null
                int newCount = it1->second - it2->second;
                assert(newCount>=0); // user can't remove a distribution that wasn't added
                if (newCount==0) {
                    it1 = this->erase(it1);
                } else {
                    it1->second = newCount;
                    ++it1;
                }
                ++it2;
            }
            total -= distribution.total;
            assert(total>=0); // user can't remove a distribution that wasn't added
            // total maybe 0, but then, the cluster will be destroyed
        }


        // Test if the 2 distributions match, perform a Chi-square computation.
        // Return a value for significance testing.
        double match(const FutureDistribution& dist) {
            // assert(dist1.total>0); assert(dist2.total>0);
            float n1 = total;
            float n2 = dist.total;
            float n1n2 = n1 * n2;
            float chisq = 0.0f;
            unsigned int ndegree = 0;

            // create a bin for each distinct entry in either of the distributions
            // use the fact that the list is sorted
            const_iterator it1 = this->begin();
            const_iterator eit1 = this->end();
            const_iterator it2 = dist.begin();
            const_iterator eit2 = dist.end();
            while (true) { // there is at least one element, assert(it1!=eit1 && it2!=eit2)
                // in each test case when applicable: it1->second!=0, it2->second!=0, by construction
                ++ndegree;
                if (it1->first < it2->first) {
                    chisq += n2 * (it1++)->second;
                    if (it1==eit1) break;
                }
                else if (it2->first < it1->first) {
                    chisq += n1 * (it2++)->second;
                    if (it2==eit2) break;
                }
                else {
                    float tmp = it1->second - it2->second;
                    chisq += n1n2 * tmp * tmp / (n1 * (it1++)->second + n2 * (it2++)->second);
                    if (it1==eit1 || it2==eit2) break;
                }
            }
            // more elements remain in list 1 ?
            while (it1!=eit1) {
                chisq += n2 * (it1++)->second;
                ++ndegree;
            }
            // more elements remain in list 2 ?
            while (it2!=eit2) {
                chisq += n1 * (it2++)->second;
                ++ndegree;
            }

            float x = 0.5f*chisq;   // constant through the recursion
            float halflogx = 0.5f*logf(x); // constant through the recursion
            float lgamma, nig; // set at lowest recurse level, log-gamma and normalized incomplete gamma
            recurse_nigamma_2a(ndegree, lgamma, nig, x, halflogx);
            if (x<1e-6f) nig = 1.0f; // igamma(a,0) = gamma(a), then normalize.
            return nig;
        }

        // Custom from-scratch implementation that is faster than usual power-series approx
        // for the special cases of this program.
        // return normalized incomplete gamma Q(0.5f*ndegree, 0.5f*chisq)
        // with: Q(a,x) = igamma(a,x) / gamma(a);
        // Iterate using the formulas:
        // - twice_a = ndegree is integer.
        // - igamma(1/2, x) = sqrt(pi).erfc(sqrt(x)), gamma(1/2) = sqrt(pi)
        // - igamma(1) = e^-x, gamma(1) = 1
        // - igamma(a, x) = (a-1)*igamma(a-1,x) + x^(a-1).e^-x  (for a>=3/2)
        // - gamma(a) = (a-1)*gamma(a-1) (for a>=3/2)
        void recurse_nigamma_2a(unsigned int twice_a, float &lgamma, float &nig, const float &x, const float &halflogx) {
            if (twice_a == 1) {
                lgamma = 0.572364942927f;
                nig = erfcf(sqrtf(x));
                return;
            }
            if (twice_a == 2) {
                lgamma = 0.0f;
                nig = expf(-x);
                return;
            }
            twice_a -= 2;
            recurse_nigamma_2a(twice_a, lgamma, nig, x, halflogx);
            lgamma += cachedLogHalf(twice_a);
            nig += expf(twice_a * halflogx - x - lgamma);
        }

        inline float cachedLogHalf(unsigned int n) {
            static std::vector<float> cachedLog;
            static std::vector<bool> hasLog;
            if (hasLog.size() > n && hasLog[n]) return cachedLog[n];
            if (hasLog.size() < n+1) {hasLog.resize(n+1); cachedLog.resize(n+1);}
            hasLog[n] = true; cachedLog[n] = logf(n) - 0.69314718056f;
            return cachedLog[n];
        }

    };

    // A cluster contains a an average future distribution,
    // mapping between pasts to cluster is external
    struct Cluster {

        // The average distribution
        FutureDistribution distribution;
        // The number of times this cluster was observed
        int ninstances;

        // cluster complexity is -log(ninstances/totalInstances);
        // keep log(ninstances) since this is only modified when changing dist
        // then log(totalInstances) is computed only once for all clusters, and
        // a simple subtraction gives the complexity (instead of a log)
        double logNinstances;

        inline Cluster(FutureDistribution* dist) : ninstances(0), prev(0), next(0) {
            distribution.cluster = this;
            merge(dist);
        }

        inline void merge(FutureDistribution* dist) {
            assert(dist->cluster == 0);
            distribution.add(*dist);
            ninstances += dist->total;
            logNinstances = log2(ninstances);
            dist->cluster = this;
        }

        inline void remove(FutureDistribution* dist) {
            distribution.sub(*dist);
            ninstances -= dist->total;
            assert(ninstances>=0); // and log = -inf doesn't matter for destroyed cluster
            logNinstances = log2(ninstances);
            dist->cluster = 0;
        }

        // maintains a double-linked list of sorted clusters
        // O(Ncluster) but OK since add/remove an observation is O(Ncluster) anyway
        Cluster* prev;
        Cluster* next;
    };

    Cluster* firstCluster;
    Cluster* lastCluster;

#ifndef NO_TR1
    typedef std::tr1::unordered_map<PastLightCone,FutureDistribution*> DistributionMap;
#else
    typedef std::map<PastLightCone,FutureDistribution*> DistributionMap;
#endif

    DistributionMap observedDistributions;

    int totalInstances;
    // subtract each cluster logNinstances from that to get that cluster complexity
    double logTotalInstances;


    typedef int (FutureDistribution::*insert_or_remove_function)(const FutureLightCone&);

    template<insert_or_remove_function insert_or_remove, int deltaObs> Cluster* add_or_remove_Observation(const PastLightCone &plc, const FutureLightCone &flc) {

        // Step 1: find distribution for that past, if it exists
        typename DistributionMap::iterator distIt = observedDistributions.find(plc);

        FutureDistribution* observedDistribution;

        Cluster* oldCluster = 0;

        // found: remove the distribution from its cluster and update both
        if (distIt != observedDistributions.end()) {

            observedDistribution = distIt->second;
            Cluster* cluster = observedDistribution->cluster;

            // update cluster distrib. This may change the ordering
            cluster->remove(observedDistribution);

            // cluster is either empty or not at its place => should be removed from list
            bool wasMoved = false;
            if (cluster->ninstances==0
            || (cluster->prev!=0 && cluster->ninstances < cluster->prev->ninstances)
            || (cluster->next!=0 && cluster->ninstances > cluster->next->ninstances)) {
                // either maintain sorted linkage, or maintain first/last elements
                if (cluster->prev) cluster->prev->next = cluster->next;
                else firstCluster = cluster->next;
                if (cluster->next) cluster->next->prev = cluster->prev;
                else lastCluster = cluster->prev;
                wasMoved = true;
            }

            // destroy the cluster if it has become empty
            if (cluster->ninstances==0) {
                delete cluster; cluster = 0;
            }
            // or re-insert it at its new place if necessary
            else if (wasMoved) {
                // re-insert it at its place
                if (!firstCluster) {
                    cluster->prev = cluster->next = 0;
                    lastCluster = firstCluster = cluster;
                }
                else if (cluster->ninstances < firstCluster->ninstances) {
                    cluster->next = firstCluster;
                    cluster->prev = 0;
                    firstCluster->prev = cluster;
                    firstCluster = cluster;
                }
                else {
                    Cluster* c = firstCluster;
                    for (; c->next!=0; c = c->next) if (cluster->ninstances < c->next->ninstances) break;
                    cluster->next = c->next;
                    cluster->prev = c;
                    if (c->next) c->next->prev = cluster;
                    else lastCluster = cluster;
                    c->next = cluster;
                }
            }

            // now update the observedDistribution
            if ((observedDistribution->*insert_or_remove)(flc) == 0) {
                // that past has no future anymore... cleanup and done!
                observedDistributions.erase(distIt);
                delete observedDistribution;
                totalInstances += deltaObs;
                logTotalInstances = log2(totalInstances);
                return cluster;
            }

            oldCluster = cluster;
        }
        // create a new distribution if this plc was not found, ensure it was found if we're removing an observation
        else {
            assert(deltaObs>0);
            observedDistribution = new FutureDistribution(flc);
            observedDistributions[plc] = observedDistribution;
        }

        // At this point, we get an updated observedDistribution (whether by adding or removing an flc)
        // Find the new cluster for that non-null observedDistribution
        Cluster* cluster = 0;

        // Step 2: run through the clusters from max complexity to low, take first match.
        // Idea: if 2 clusters pass the stat. test, take the more complex one.
        // this forces high-complexity clusters to be really different hence enhance the contrast.
        // conversely the low-complexity are better merged as their elements get reevaluated after the initial phase.
        bool shouldTestOldCluster = false;
        for (Cluster* c = firstCluster; c!=0; c = c->next) {
            if (c == oldCluster) {shouldTestOldCluster = true; continue;}
            if (observedDistribution->match(c->distribution) < significanceLevel) {
                cluster = c;
                break;
            }
        }
        if ((cluster==0) && shouldTestOldCluster) {
            if (observedDistribution->match(oldCluster->distribution) < significanceLevel) {
                cluster = oldCluster;
            }
        }

        // Step 3: Insert the distribution in the cluster just found, or create a new
        // one. This may be or not the old cluster. Update the cluster position in the
        // sorted cluster set.

        // no cluster found, create a new one
        bool wasMoved = false;
        if (cluster == 0) {
            cluster = new Cluster(observedDistribution);
            wasMoved = true;
        }
        // or update it
        else {
            cluster->merge(observedDistribution);
            // remove the cluster from the sorted list if its position changed
            if ((cluster->prev!=0 && cluster->ninstances < cluster->prev->ninstances)
             || (cluster->next!=0 && cluster->ninstances > cluster->next->ninstances)) {
                // either maintain sorted linkage, or maintain first/last elements
                if (cluster->prev) cluster->prev->next = cluster->next;
                else firstCluster = cluster->next;
                if (cluster->next) cluster->next->prev = cluster->prev;
                else lastCluster = cluster->prev;
                wasMoved = true;
            }
        }

        // insert the cluster at its new place if necessary
        if (wasMoved) {
            // re-insert it at its place
            if (!firstCluster) {
                cluster->prev = cluster->next = 0;
                lastCluster = firstCluster = cluster;
            }
            else if (cluster->ninstances < firstCluster->ninstances) {
                cluster->next = firstCluster;
                cluster->prev = 0;
                firstCluster->prev = cluster;
                firstCluster = cluster;
            }
            else {
                Cluster* c = firstCluster;
                for (; c->next!=0; c = c->next) if (cluster->ninstances < c->next->ninstances) break;
                cluster->next = c->next;
                cluster->prev = c;
                if (c->next) c->next->prev = cluster;
                else lastCluster = cluster;
                c->next = cluster;
            }
        }

        // update total & log
        totalInstances += deltaObs;
        logTotalInstances = log2(totalInstances);

        return cluster;
    }


    inline ~CausalStateAnalyzer() {
        while (firstCluster!=0) {
            Cluster* next = firstCluster->next;
            delete firstCluster;
            firstCluster = next;
        }
    }



// USER API

    // Create an Analyzer
    // the future distributions will be clustered when they match below the given significance level
    // Default is 0.04321, just so it's not 0.05.
    inline CausalStateAnalyzer(double _significanceLevel = 0.04321)
    : significanceLevel(_significanceLevel), firstCluster(0), lastCluster(0), totalInstances(0) {
    }

    // Add a new observed future light cone for a given past light cone
    // This will update the plc distibution and the clusters and their complexity
    // The (potentially new) cluster for that plc is returned.
    inline Cluster* addObservation(const PastLightCone& plc, const FutureLightCone& flc) {
        return add_or_remove_Observation<&FutureDistribution::insert,1>(plc,flc);
    }

    // Useful for systems changing over time or space. When "old" observations
    // have a different enough distribution from current observation (non-stationary system)
    // and should be discarded.
    // If there are still instances for that plc, the corresponding cluster is returned, otherwise 0.
    inline Cluster* removeObservation(const PastLightCone& plc, const FutureLightCone& flc) {
        return add_or_remove_Observation<&FutureDistribution::remove,-1>(plc,flc);
    }

    // get absolute cluster complexity
    inline double getComplexity(Cluster* cluster) {
        return logTotalInstances - cluster->logNinstances;
    }

    // get cluster complexity scaled to 0..1 relatively to the other clusters
    inline double getScaledComplexity(Cluster* cluster) {
        double minLog = firstCluster->logNinstances;
        double maxLog = lastCluster->logNinstances;
        double scaleFactor = maxLog - minLog;
        if (scaleFactor!=0.0) scaleFactor = 1.0 / scaleFactor;
        return (maxLog - cluster->logNinstances) * scaleFactor;
    }

    // Get the global complexity of the analyzer
    inline double getGlobalComplexity() {
        // This could be made incremental
        double globalComplexity = 0.0;
        for (Cluster* c = firstCluster; c!=0; c = c->next) {
            globalComplexity += c->ninstances * c->logNinstances;
        }
        return logTotalInstances - globalComplexity / totalInstances;
    }

};

